# Version 0.0.6

# changes for DXF ouput Thomas Lachmann 2013
# - DXF output instead of SVG
# - color mapping to ACI colors
# - rendersize calculated
# - frustum culling added
# 2013-12-10
# - parameters from dxfWRiterGUI.py added for controlling the generated edges
# 2013-12-14
#  - no use of dxfLibrary any more
#  - write dxf R14 files with truecolor support
#  - implemented a color shader which takes the original material color
# 2014-03-16
#  - additional option to get colors from rendered image
# 2014-03-24
#  - support for Blender 2.70
#  - additional color renderer
# 2014-04-19
#  -  chaining support
#  -  some code cleanup

# based on work from Tamito KAJIYAMA <19 August 2009>

import bpy
import os
import math
from mathutils import *
from parameter_editor import *
from dxfFile import *

from freestyle.chainingiterators import *
from freestyle.functions import *
from freestyle.predicates import (
    NotUP1D,
    QuantitativeInvisibilityUP1D,
    pyDensityUP1D,
    pyHigherLengthUP1D,
    pyHigherNumberOfTurnsUP1D,
    pyLengthBP1D,
    pyParameterUP0D,
    )
from freestyle.shaders import (
    BezierCurveShader,
    ConstantColorShader,
    ConstantThicknessShader,
    SamplingShader,
    TextureAssignerShader,
    TipRemoverShader,
    pyNonLinearVaryingThicknessShader,
    pySamplingShader,
    )
from freestyle.types import IntegrationType, Operators
import freestyle

global img_pixels
global img_width

CLIP_LEFT = 1
CLIP_RIGHT = 2
CLIP_TOP = 3
CLIP_BOTTOM = 4

### T.K. 07-Aug-2013 Temporary fix for unexpected line gaps

def iter_three_segments(stroke):
    n = stroke.stroke_vertices_size()
    if n >= 4:
        it1 = stroke.stroke_vertices_begin()
        it2 = stroke.stroke_vertices_begin()
        it2.increment()
        it3 = stroke.stroke_vertices_begin()
        it3.increment()
        it3.increment()
        it4 = stroke.stroke_vertices_begin()
        it4.increment()
        it4.increment()
        it4.increment()
        while not it4.is_end:
            yield (it1.object, it2.object, it3.object, it4.object)
            it1.increment()
            it2.increment()
            it3.increment()
            it4.increment()


def is_tvertex(svertex):
    return type(svertex.viewvertex) is TVertex


class StrokeCleaner(StrokeShader):
    def shade(self, stroke):
        for sv1, sv2, sv3, sv4 in iter_three_segments(stroke):
            seg1 = sv2.point - sv1.point
            seg2 = sv3.point - sv2.point
            seg3 = sv4.point - sv3.point
            if not ((is_tvertex(sv2.first_svertex) and is_tvertex(sv2.second_svertex)) or
                    (is_tvertex(sv3.first_svertex) and is_tvertex(sv3.second_svertex))):
                continue
            if seg1.dot(seg2) < 0.0 and seg2.dot(seg3) < 0.0 and seg2.length < 0.01:
                #print(sv2.first_svertex.viewvertex)
                #print(sv2.second_svertex.viewvertex)
                #print(sv3.first_svertex.viewvertex)
                #print(sv3.second_svertex.viewvertex)
                p2 = mathutils.Vector(sv2.point)
                p3 = mathutils.Vector(sv3.point)
                sv2.point = p3
                sv3.point = p2
        stroke.update_length()


def PixelGetRGBA(x,y):
    offset = (y*img_width + x)*4
    return img_pixels[offset:offset+4]


class myMaterialColorShader(StrokeShader):
    def __init__(self):
        StrokeShader.__init__(self)
    def shade(self, stroke):
        it = stroke.stroke_vertices_begin()
        func = MaterialF0D()
        while not it.is_end:
            mat = func(Interface0DIterator(it))
            r = mat.diffuse[0]
            g = mat.diffuse[1]
            b = mat.diffuse[2]
            it.object.attribute.color = (r, g, b)
            it.increment()


class myRenderColorShader(StrokeShader):
    def __init__(self):
        StrokeShader.__init__(self)
    def shade(self, stroke):
        it = stroke.stroke_vertices_begin()
        last_object = it.object
        if not it.is_end:
            it.increment()
        while not it.is_end:
            #print("Color Shader %.3f %.3f" % (it.object.point.x, it.object.point.y))
            #print("Color Shader %.3f %.3f" % ((last_object.point.x + it.object.point.x) / 2, (last_object.point.y + it.object.point.y) / 2))
            color = PixelGetRGBA(math.trunc((last_object.point.x + it.object.point.x) / 2), math.trunc((last_object.point.y + it.object.point.y) / 2))	# pick color in the middle of the stroke
            #color = PixelGetRGBA(math.trunc((math.trunc(last_object.point.x) + math.trunc(it.object.point.x)) / 2), math.trunc((math.trunc(last_object.point.y) + math.trunc(it.object.point.y)) / 2))	# pick color in the middle of the stroke
            #color = PixelGetRGBA(math.trunc(it.object.point.x), math.trunc(it.object.point.y))	# pick color directly from rendered image @ x y		
            it.object.attribute.color = (color[0], color[1], color[2])	# RGB
            last_object.attribute.color = (color[0], color[1], color[2])	# RGB
            last_object = it.object
            it.increment()


class DXFWriter(StrokeShader):
    def __init__(self, f, w, h):
        StrokeShader.__init__(self)
        self.width, self.height = w, h
        # create the file 
        self.file = f
        if self.file != None:
            self.file.write(DXF_HEADER % (w, h))
        self.dxfHandle = 100
        
    def close(self):
        try:
            if self.file != None:
                self.file.write(DXF_FOOTER)
                self.file.close()
                print("dxf saved to file %r" % self.file.name)
        except:
            print("Cannot save to file")

    def isPointOutsideViewport(self, point): 
        x, y = point
        if x < 0:
            return CLIP_LEFT
        if y < 0:
            return CLIP_BOTTOM
        if x > self.width:
            return CLIP_RIGHT
        if y > self.height:
            return CLIP_TOP
        return None
            
    def shade(self, stroke):

        points = []
        for v in stroke:
            x, y = v.point
            points.append((x, y))
        
        points_clipped = []	
        if len(stroke) > 0:
            last_clip = self.isPointOutsideViewport(points[0])
            if last_clip == None:
                points_clipped.append(DXF_VERTEX % (points[0][0], points[0][1]))

        for i in range(1, len(points)):
            clip = self.isPointOutsideViewport(points[i])
            if (clip == None) and (last_clip == None):
                points_clipped.append(DXF_VERTEX % (points[i][0], points[i][1]))
            else:
                clip_tmp = last_clip
                if (last_clip == None):
                    clip_tmp = clip
                if clip_tmp == CLIP_LEFT:
                    clip_p1 = (0,0)
                    clip_p2 = (0,self.height)
                elif clip_tmp == CLIP_RIGHT:
                    clip_p1 = (self.width,0)
                    clip_p2 = (self.width,self.height)
                elif clip_tmp == CLIP_TOP:
                    clip_p1 = (0,self.height)
                    clip_p2 = (self.width,self.height)
                elif clip_tmp == CLIP_BOTTOM:
                    clip_p1 = (0,0)
                    clip_p2 = (self.width,0)
                intersect = mathutils.geometry.intersect_line_line_2d(mathutils.Vector(clip_p1), mathutils.Vector(clip_p2), mathutils.Vector(points[i-1]), mathutils.Vector(points[i]))
                if intersect:
                    points_clipped.append(DXF_VERTEX % (intersect.x, intersect.y))
                    if last_clip != None:
                        points_clipped.append(DXF_VERTEX % (points[i][0], points[i][1]))
                last_clip = clip
        
        if len(points_clipped) > 0:
            if self.file != None:
                self.file.write(DXF_LWPOLYLINE % (self.dxfHandle, dxfColor(v.attribute.color), len(points_clipped), "".join(points_clipped)))
            self.dxfHandle += 1


scene = freestyle.utils.getCurrentScene()

output_dir = bpy.path.abspath(scene.render.filepath)
if not os.path.exists(output_dir):
    os.makedirs(output_dir)
path = os.path.join(output_dir, "Frame%06d.dxf" % scene.frame_current)
try:
    f = open(path, "wt")
    file_error = False
except:
    print("Cannot save to file %r" % path)
    f = None

image_path = None

w = scene.render.resolution_x * scene.render.resolution_percentage / 100
h = scene.render.resolution_y * scene.render.resolution_percentage / 100

coloring = 'MATERIAL'

if  hasattr(scene, 'dxfWriter_settings'):
    print('dxfWriter settings found.')
    lineset = scene.dxfWriter_settings
    if lineset.coloring == 'RENDERED':
        #image_path = os.path.join(output_dir, "tmp.tmp")
        image_path = "tmp.tmp"
        bpy.data.images['Render Result'].save_render(image_path)
        img = bpy.data.images.load(image_path)
        img_pixels = img.pixels[:]
        img_width = img.size[0]

    selection_criteria = []
    # prepare selection criteria by visibility
    if lineset.select_by_visibility:
        if lineset.visibility == 'VISIBLE':
            selection_criteria.append(
                QuantitativeInvisibilityUP1D(0))
        elif lineset.visibility == 'HIDDEN':
            selection_criteria.append(
                NotUP1D(QuantitativeInvisibilityUP1D(0)))
        elif lineset.visibility == 'RANGE':
            selection_criteria.append(
                QuantitativeInvisibilityRangeUP1D(lineset.qi_start, lineset.qi_end))
    # prepare selection criteria by edge types
    if lineset.select_by_edge_types:
        edge_type_criteria = []
        if lineset.select_silhouette:
            upred = pyNatureUP1D(Nature.SILHOUETTE)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_silhouette else upred)
        if lineset.select_border:
            upred = pyNatureUP1D(Nature.BORDER)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_border else upred)
        if lineset.select_crease:
            upred = pyNatureUP1D(Nature.CREASE)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_crease else upred)
        if lineset.select_ridge_valley:
            upred = pyNatureUP1D(Nature.RIDGE)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_ridge_valley else upred)
        if lineset.select_suggestive_contour:
            upred = pyNatureUP1D(Nature.SUGGESTIVE_CONTOUR)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_suggestive_contour else upred)
        if lineset.select_material_boundary:
            upred = pyNatureUP1D(Nature.MATERIAL_BOUNDARY)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_material_boundary else upred)
        if lineset.select_edge_mark:
            upred = pyNatureUP1D(Nature.EDGE_MARK)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_edge_mark else upred)
        if lineset.select_contour:
            upred = ContourUP1D()
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_contour else upred)
        if lineset.select_external_contour:
            upred = ExternalContourUP1D()
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_external_contour else upred)
        if lineset.edge_type_combination == 'OR':
            upred = join_unary_predicates(edge_type_criteria, OrUP1D)
        else:
            upred = join_unary_predicates(edge_type_criteria, AndUP1D)
        if upred is not None:
            if lineset.edge_type_negation == 'EXCLUSIVE':
                upred = NotUP1D(upred)
            selection_criteria.append(upred)
    # prepare selection criteria by face marks
    if lineset.select_by_face_marks:
        if lineset.face_mark_condition == 'BOTH':
            upred = FaceMarkBothUP1D()
        else:
            upred = FaceMarkOneUP1D()
        if lineset.face_mark_negation == 'EXCLUSIVE':
            upred = NotUP1D(upred)
        selection_criteria.append(upred)
    # prepare selection criteria by group of objects
    #~ if lineset.select_by_group:
        #~ if lineset.group is not None:
            #~ names = dict((ob.name, True) for ob in lineset.group.objects)
            #~ upred = ObjectNamesUP1D(names, lineset.group_negation == 'EXCLUSIVE')
            #~ selection_criteria.append(upred)
    # prepare selection criteria by image border
    if lineset.select_by_image_border:
        xmin, ymin, xmax, ymax = ContextFunctions.get_border()
        upred = WithinImageBoundaryUP1D(xmin, ymin, xmax, ymax)
        selection_criteria.append(upred)
    # select feature edges
    upred = join_unary_predicates(selection_criteria, AndUP1D)
    if upred is None:
        upred = TrueUP1D()
    # coloring
    coloring = lineset.coloring
else:
    upred = QuantitativeInvisibilityUP1D(0)
    
Operators.select(upred)

#Operators.bidirectionalChain(ChainSilhouetteIterator(), NotUP1D(upred))

# chaining
if lineset.use_chaining:
		if lineset.use_same_object:
			Operators.bidirectional_chain(ChainSilhouetteIterator(), NotUP1D(upred))
		else:
			Operators.bidirectional_chain(ChainPredicateIterator(upred, TrueBP1D()), NotUP1D(upred))
else:
		Operators.chain(ChainPredicateIterator(FalseUP1D(), FalseBP1D()), NotUP1D(upred))

#  split  at material boundaries
Operators.sequential_split(MaterialBoundaryUP0D())

writer = DXFWriter(f, w, h)

shaders_list = []
shaders_list.append(StrokeCleaner())
shaders_list.append(ConstantThicknessShader(2))
if coloring == 'RENDERED':
	shaders_list.append(myRenderColorShader())
else:
	shaders_list.append(myMaterialColorShader())
shaders_list.append(writer)
    
Operators.create(TrueUP1D(), shaders_list)

if  image_path != None:
    img.user_clear()
    bpy.data.images.remove(img)
    os.remove(image_path)

writer.close()

